from .base import Application, Handler, Resource, ResourceConnector
from .connectors import S3Permission, SqsPermission, s3, sqs
from .handlers import (
    JobEvent,
    QueueEvent,
    Request,
    Response,
    S3EventSequence,
    SqsEventsSequence,
    SqsFailedEvents,
    job,
    route,
    s3_handler,
    sqs_handler,
)
from .resources import S3, Sqs
from .utils import handler_config, inject_variables



api_template = """from viburnum.application import route, Request, Response


@route("{path}", methods={methods})
def {handler_name}(request: Request) -> Response:
    return Response(200, {{}})

"""



import enum
import json
from collections import UserList
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, Iterable, Optional

from viburnum.application.base import Handler, LambdaInput, LambdaOutput
from viburnum.application.types import HeadersType, JsonData, MultiQueryParamsType

# ___________________ API __________________________
# Lambda with Rest Api
# https://docs.aws.amazon.com/lambda/latest/dg/services-apigateway.html#apigateway-example-event


class Request(LambdaInput):
    def __init__(self, event: dict, context: dict) -> None:
        super().__init__(event, context)
        self._json = None

    @property
    def raw_headers(self) -> HeadersType:
        return self.event["headers"]

    @property
    def raw_query_params(self) -> MultiQueryParamsType:
        return self.event["multiValueQueryStringParameters"]

    @property
    def method(self) -> str:
        return self.event["httpMethod"]

    @property
    def path(self) -> str:
        return self.event["path"]

    @property
    def path_params(self) -> dict:
        return self.event["pathParameters"] or {}

    @property
    def body(self) -> Optional[str]:
        return self.event["body"]

    def json(self) -> Optional[JsonData]:
        if not self.body:
            return None
        if not self._json:
            self._json = json.loads(self.body)
        return self._json


class Response(LambdaOutput):
    def __init__(
        self,
        status_code: int,
        body: dict,
        headers: Optional[dict] = None,
        is_base_64: bool = False,
    ) -> None:
        # TODO: add support for multivalue headers
        self.response_data = {
            "statusCode": status_code,
            "headers": headers or {},
            "isBase64Encoded": is_base_64,
            "body": json.dumps(body),
        }

    def as_response(self) -> dict:
        return self.response_data


class ApiHandler(Handler):
    event_class = Request

    def __init__(
        self,
        func: Callable,
        path: str,
        methods: Iterable[str],
    ) -> None:
        self.path: str = path
        self.methods: Iterable[str] = methods
        super().__init__(func)

    @staticmethod
    def _name_suffix() -> str:
        return "_api"


def route(path: str, methods: Iterable[str] = ("ANY",)):
    "Wrapper for creating :class:`ApiHandler` resource."

    def wraper(func):
        return ApiHandler(func, path, methods)

    return wraper


# ___________________ Job ______________________________
# Lambda with EventBridge
# https://docs.aws.amazon.com/lambda/latest/dg/services-cloudwatchevents.html


class JobEvent(LambdaInput):
    @property
    def detail(self) -> dict[str, Any]:
        return self.event["detail"]


class JobHandler(Handler):
    event_class = JobEvent

    def __init__(self, func: Callable, schedule: str) -> None:
        self.schedule = schedule
        super().__init__(func)

    @staticmethod
    def _name_suffix() -> str:
        return "_job"


def job(schedule: str):
    """
    Wrapper for creating :class:`JobHandler` resource.
    Schedule expression [docs](https://docs.aws.amazon.com/eventbridge/latest/userguide/eb-create-rule-schedule.html)
    """

    def wrapper(func):
        return JobHandler(func, schedule)

    return wrapper


# __________________ Sqs Worker ____________________________
# Lambda with SQS
# https://docs.aws.amazon.com/lambda/latest/dg/with-sqs.html


class QueueEvent:
    def __init__(self, event: dict) -> None:
        self.event = event
        self._body = None

    def delete(self):
        pass
        # TODO: delete event from queue

    @property
    def message_id(self) -> str:
        return self.event["messageId"]

    @property
    def body(self) -> JsonData:
        if self._body is None:
            try:
                self._body = json.loads(self.event["body"])
            except json.JSONDecodeError:
                self._body = self.event["body"]
        return self._body

    @property
    def attributes(self) -> dict[str, Any]:
        return self.event["attributes"]

    @property
    def message_attributes(self) -> dict[str, Any]:
        return self.event["messageAttributes"]


class SqsEventsSequence(LambdaInput, UserList[QueueEvent]):
    def __init__(self, event: dict, context: dict) -> None:
        super().__init__(event, context)
        self.data = [QueueEvent(e) for e in self.event["Records"]]


class SqsFailedEvents:
    # Returns from lambda failed events
    # https://docs.aws.amazon.com/lambda/latest/dg/with-sqs.html#services-sqs-batchfailurereporting

    def __init__(self, *failed_ids: str) -> None:
        self.failed_event_ids: set[str] = set(failed_ids)

    def fail(self, *failed_ids: str):
        self.failed_event_ids.update(failed_ids)

    def as_response(self) -> dict:
        return {
            "batchItemFailures": [
                {"itemIdentifier": id for id in self.failed_event_ids}
            ]
        }


class SqsHandler(Handler):
    event_class = SqsEventsSequence

    @staticmethod
    def _name_suffix() -> str:
        return "_worker"

    def __init__(self, func: Callable, queue_name: str) -> None:
        self.queue_name = queue_name
        super().__init__(func)


def sqs_handler(queue_name: str):
    """
    Wrapper for creating :class:`JobHandler` resource.
    """

    def wrapper(func):
        return SqsHandler(
            func,
            queue_name,
        )

    return wrapper


# _________________ S3 Worker ____________________________
# Lambda with S3 stream
# https://docs.aws.amazon.com/lambda/latest/dg/with-s3.html


class S3EventType(enum.Enum):
    """Notification event types.

    :link: https://docs.aws.amazon.com/AmazonS3/latest/userguide/notification-how-to-event-types-and-destinations.html#supported-notification-event-types
    """

    OBJECT_CREATED = "OBJECT_CREATED"
    OBJECT_CREATED_PUT = "OBJECT_CREATED_PUT"
    OBJECT_CREATED_POST = "OBJECT_CREATED_POST"
    OBJECT_CREATED_COPY = "OBJECT_CREATED_COPY"
    OBJECT_CREATED_COMPLETE_MULTIPART_UPLOAD = (
        "OBJECT_CREATED_COMPLETE_MULTIPART_UPLOAD"
    )
    OBJECT_REMOVED = "OBJECT_REMOVED"
    OBJECT_REMOVED_DELETE = "OBJECT_REMOVED_DELETE"
    OBJECT_REMOVED_DELETE_MARKER_CREATED = "OBJECT_REMOVED_DELETE_MARKER_CREATED"
    OBJECT_RESTORE_POST = "OBJECT_RESTORE_POST"
    OBJECT_RESTORE_COMPLETED = "OBJECT_RESTORE_COMPLETED"
    OBJECT_RESTORE_DELETE = "OBJECT_RESTORE_DELETE"
    REDUCED_REDUNDANCY_LOST_OBJECT = "REDUCED_REDUNDANCY_LOST_OBJECT"
    REPLICATION_OPERATION_FAILED_REPLICATION = (
        "REPLICATION_OPERATION_FAILED_REPLICATION"
    )
    REPLICATION_OPERATION_MISSED_THRESHOLD = "REPLICATION_OPERATION_MISSED_THRESHOLD"
    REPLICATION_OPERATION_REPLICATED_AFTER_THRESHOLD = (
        "REPLICATION_OPERATION_REPLICATED_AFTER_THRESHOLD"
    )
    REPLICATION_OPERATION_NOT_TRACKED = "REPLICATION_OPERATION_NOT_TRACKED"
    LIFECYCLE_EXPIRATION = "LIFECYCLE_EXPIRATION"
    LIFECYCLE_EXPIRATION_DELETE = "LIFECYCLE_EXPIRATION_DELETE"
    LIFECYCLE_EXPIRATION_DELETE_MARKER_CREATED = (
        "LIFECYCLE_EXPIRATION_DELETE_MARKER_CREATED"
    )
    LIFECYCLE_TRANSITION = "LIFECYCLE_TRANSITION"
    INTELLIGENT_TIERING = "INTELLIGENT_TIERING"
    OBJECT_TAGGING = "OBJECT_TAGGING"
    OBJECT_TAGGING_PUT = "OBJECT_TAGGING_PUT"
    OBJECT_TAGGING_DELETE = "OBJECT_TAGGING_DELETE"
    OBJECT_ACL_PUT = "OBJECT_ACL_PUT"


@dataclass
class S3Object:
    key: str
    size: int
    eTag: str
    sequencer: str


@dataclass
class S3Bucket:
    name: str
    arn: str


@dataclass
class S3Event:
    event_name: str
    event_time: datetime
    bucket: S3Bucket
    object: S3Object


class S3EventSequence(LambdaInput, UserList[S3Event]):
    def __init__(self, event: dict, context: dict) -> None:
        super().__init__(event, context)
        self.data = [self._get_s3_event(e) for e in event["Records"]]

    def _get_s3_event(self, raw_event: dict) -> S3Event:
        bucket = S3Bucket(
            name=raw_event["s3"]["bucket"]["name"], arn=raw_event["s3"]["bucket"]["arn"]
        )
        object_ = S3Object(**raw_event["s3"]["object"])
        return S3Event(
            event_name=raw_event["eventName"],
            event_time=datetime.strptime(
                raw_event["eventTime"], "%Y-%m-%dT%H:%M:%S.%fZ"
            ),
            bucket=bucket,
            object=object_,
        )


class S3Handler(Handler):
    event_class = S3EventSequence

    def __init__(
        self, func: Callable, bucket_name: str, events: Iterable[S3EventType]
    ) -> None:
        super().__init__(func)
        self.bucket_name = bucket_name
        self.events = events


def s3_handler(
    bucket_name: str, events: Iterable[S3EventType] = (S3EventType.OBJECT_CREATED,)
):
    """
    Wrapper for creating :class:`JobHandler` resource.
    """

    def wrapper(func):
        return S3Handler(
            func,
            bucket_name,
            events,
        )

    return wrapper



from typing import Any, Dict, Optional

from viburnum.application.base import Handler


def handler_config(
    memory: Optional[int] = None,
    timeout: Optional[int] = None,
    reserved_concurrent_executions: Optional[int] = None,
):
    """Set configuration for specific lambda.

    Args:
        memory (Optional[int], optional): memory size in MB. Defaults to value provided in Configuration.
        timeout (Optional[int], optional): lambda timeout in seconds. Defaults to value provided in Configuration.
        reserved_concurrent_executions (Optional[int], optional): number of reserved concurrent executions. Defaults to value provided in Configuration.
    """

    def wraper(handler: Handler) -> Handler:
        handler.memory = memory
        handler.timeout = timeout
        handler.reserved_concurrent_executions = reserved_concurrent_executions
        return handler

    return wraper


def inject_variables(env_variables: Dict[str, Any]):
    """Inject environment variables for specific lambda.

    Args:
        env_variables (Dict[str, Any]): dict with variables
    """

    def wraper(handler: Handler) -> Handler:
        handler.env_variables.update(env_variables)
        return handler

    return wraper



import os
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, Optional, Sequence

from aws_cdk import Environment, aws_lambda


@dataclass(frozen=True)
class LambdaDefaults:
    memory_size: int = 128
    timeout: int = 25
    reserved_concurrent_executions: Optional[int] = None
    env_variables: Dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True)
class Configuration:
    env_name: str
    env: Environment  # see https://docs.aws.amazon.com/cdk/v2/guide/environments.html
    python_version: aws_lambda.Runtime
    # lambda config
    lambda_defaults: LambdaDefaults = LambdaDefaults()
    # vpc config
    vpc_id: Optional[str] = None
    security_group_id: Optional[str] = None
    subnets_ids: Optional[Sequence[str]] = None
    allow_public_subnet: bool = False

    __configs__: ClassVar[Dict[str, "Configuration"]] = {}

    def __post_init__(self) -> None:
        if self.env_name.lower() in self.__configs__:
            raise ValueError(
                f"Configuration with env_name '{self.env_name}' already exists."
            )
        self.__configs__[self.env_name.lower()] = self
        self.__validate_vpc_config()

    def __validate_vpc_config(self) -> None:
        if self.vpc_id:
            assert self.security_group_id, "security_group_id not provided"
            assert self.subnets_ids, "subnets_ids not provided"

    @classmethod
    def get_configuration(cls) -> "Configuration":
        __import__("config")  # hack to import configurations from config.py
        env_name = os.getenv("ENV", "")
        config = cls.__configs__.get(env_name.lower())  # type: ignore
        if not config:
            raise ValueError(f"No config with env_name='{env_name}'.")
        return config



import enum
from pathlib import Path

import typer

from viburnum import __version__

from .api_template import api_template
from .app_template import app_template
from .job_template import job_template
from .s3_worker_template import s3_worker_template
from .sqs_worker_template import sqs_worker_template


def _create_moudle(path: Path):
    if not path.exists():
        path.mkdir(0o770, True, False)
    init_file = path.joinpath("__init__.py")
    if not init_file.exists():
        init_file.touch(0o770, True)


class ProjectInitializer:
    def __init__(self, without_shared: bool = False) -> None:
        typer.confirm("It will overwrite app.py file, continue?", abort=True)
        self.app_name = self._get_app_name()
        typer.secho("Initializing project...", fg=typer.colors.BRIGHT_CYAN)
        self._create_requirements()
        self._update_app()
        if not without_shared:
            self._create_shared_folder()
        typer.secho("Done!", fg=typer.colors.BRIGHT_CYAN)

    def _create_requirements(self) -> None:
        with open("requirements.txt", "w", encoding="utf-8") as f:
            f.write(f"viburnum>={__version__}")

    def _get_app_name(self) -> str:
        return typer.prompt("Stack name", type=str)

    def _update_app(self):
        with open(Path("./app.py"), "w", encoding="utf-8") as f:
            f.write(app_template.format(app_name=self.app_name))

    def _create_shared_folder(self):
        path = Path("./shared")
        _create_moudle(path)


app = typer.Typer(name="viburnum-cli")


@app.command(help="Initialize project")
def init(without_shared: bool = typer.Option(False, help="Don't create shared folder")):
    ProjectInitializer(without_shared)


class HandlerCreator:
    handler_type: str = "other"
    handler_file_template: str = "dummy"

    def __init__(self) -> None:
        self._read_params()
        self._create_handler_folder()
        self._create_handler_file()
        self._add_handler_into_app()

    def _read_params(self):
        # TODO: add regex validation
        self.handler_name = (
            typer.prompt("Function name", type=str)
            .strip()
            .replace(" ", "_")
            .replace("-", "_")
        )
        custom_path = self._get_handler_path()

        self.handler_path = Path(
            "./functions", self.handler_type, custom_path, self.handler_name
        )

    @staticmethod
    def _get_handler_path() -> str:
        path: str = typer.prompt(
            "Path where place handler",
            default="/",
        )
        return path.strip("/")

    def _create_handler_folder(self):
        _create_moudle(self.handler_path)
        for p in self.handler_path.parents[:-1]:
            _create_moudle(p)

    def _get_params(self) -> dict:
        """Overload this methods in child classes"""
        return {"handler_name": self.handler_name}

    def _create_handler_file(self):
        with open(self.handler_path.joinpath("handler.py"), "x", encoding="utf-8") as f:
            f.write(self.handler_file_template.format(**self._get_params()))

    def _add_handler_into_app(self):
        file_data = ""
        with open(Path("./app.py"), "r", encoding="utf-8") as f:
            file_data = f.read()
        with open(Path("./app.py"), "w", encoding="utf-8") as f:
            import_path = f"{self.handler_path.as_posix().replace('/', '.')}.handler"
            f.write(
                file_data.replace(
                    "# [Handlers]",
                    f"# [Handlers]\napp.add_handler({self.handler_name})",
                ).replace(
                    "# [Imports]",
                    f"# [Imports]\nfrom {import_path} import {self.handler_name}",
                )
            )


create_handler_app = typer.Typer()


ALLOWED_METHODS = ["ANY", "GET", "POST", "PUT", "PATCH", "DELETE"]


class HttpMethods(str, enum.Enum):
    ANY = "ANY"
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


def _validate_methods(value: str):
    methods = value.upper().strip().split()
    if isinstance(methods, str):
        return [HttpMethods(methods).value]
    return [HttpMethods(m).value for m in methods if m]


class ApiHandlerCreator(HandlerCreator):
    handler_type: str = "api"
    handler_file_template: str = api_template

    def __init__(self) -> None:
        super().__init__()
        typer.secho(
            f"Created api handler '{self.handler_name}'\nRoute: {self.methods} {self.path}",
            fg=typer.colors.BRIGHT_GREEN,
        )

    def _read_params(self):
        super()._read_params()
        self.methods = self._get_methods()
        self.path: str = typer.prompt("Path").strip().replace(" ", "-")
        if not self.path.startswith("/"):
            self.path = f"/{self.path}"

    def _get_methods(self) -> list[str]:
        while True:
            try:
                methods: list[str] = typer.prompt(
                    "Methods",
                    default="ANY",
                    type=HttpMethods,
                    value_proc=_validate_methods,
                    show_choices=True,
                )
                return methods
            except ValueError as e:
                typer.secho(f"Wrong value {e.args[0]}", fg=typer.colors.RED)

    def _get_params(self) -> dict:
        params = super()._get_params()
        params.update({"methods": self.methods, "path": self.path})
        return params


class JobHandlerCreator(HandlerCreator):
    handler_type: str = "jobs"
    handler_file_template: str = job_template

    def __init__(self) -> None:
        super().__init__()
        typer.secho(
            f"Created job handler '{self.handler_name}'", fg=typer.colors.BRIGHT_GREEN
        )

    def _read_params(self):
        super()._read_params()
        self.schedule = typer.prompt("Schedule expression", type=str)

    def _get_params(self) -> dict:
        params = super()._get_params()
        params.update({"schedule": self.schedule})
        return params


class SqsWorkerHandlerCreator(HandlerCreator):
    handler_type: str = "workers"
    handler_file_template: str = sqs_worker_template

    def __init__(self) -> None:
        super().__init__()
        typer.secho(
            f"Created sqs handler '{self.handler_name}'", fg=typer.colors.BRIGHT_GREEN
        )

    def _read_params(self):
        super()._read_params()
        self.sqs_name = typer.prompt("Sqs name", type=str)

    def _get_params(self) -> dict:
        params = super()._get_params()
        params.update({"sqs_name": self.sqs_name})
        return params


class S3WorkerHandlerCreator(HandlerCreator):
    handler_type: str = "workers"
    handler_file_template: str = s3_worker_template

    def __init__(self) -> None:
        super().__init__()
        typer.secho(
            f"Created s3 handler '{self.handler_name}'", fg=typer.colors.BRIGHT_GREEN
        )

    def _read_params(self):
        super()._read_params()
        self.bucket_name = typer.prompt("S3 bucket name", type=str)

    def _get_params(self) -> dict:
        params = super()._get_params()
        params.update({"bucket_name": self.bucket_name})
        return params


@create_handler_app.command()
def api():
    ApiHandlerCreator()


@create_handler_app.command()
def job():
    JobHandlerCreator()


@create_handler_app.command()
def sqs_worker():
    SqsWorkerHandlerCreator()


@create_handler_app.command()
def s3_worker():
    S3WorkerHandlerCreator()


app.add_typer(create_handler_app, name="add")


if __name__ == "__main__":
    app()



s3_worker_template = """from viburnum.application import s3_handler, S3EventSequence


@s3_handler("{bucket_name}")
def {handler_name}(events: S3EventSequence):
    for e in events:
        print(e.object.key)
"""



import inspect
import logging
import os
import shutil
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Generic, TypeVar

from aws_cdk import Duration
from aws_cdk import Resource as AwsResource
from aws_cdk import (
    Stack,
    aws_apigateway,
    aws_ec2,
    aws_events,
    aws_events_targets,
    aws_lambda,
    aws_lambda_event_sources,
    aws_lambda_python_alpha,
    aws_s3,
    aws_sqs,
)
from constructs import Construct

from viburnum.application import (
    S3,
    Application,
    Handler,
    Resource,
    ResourceConnector,
    S3Permission,
    Sqs,
    SqsPermission,
)
from viburnum.application.connectors import S3Connector, SqsConnector
from viburnum.application.handlers import ApiHandler, JobHandler, S3Handler, SqsHandler
from viburnum.deployer.configuration import Configuration


class BuilderException(Exception):
    pass


class ResourceAlreadyRegistered(BuilderException):
    def __init__(self, resource_name: str) -> None:
        self.resource_name = resource_name
        super().__init__(f"Resource '{resource_name}' already defined in stack!")


class ResourceNotDefined(BuilderException):
    def __init__(self, resource_name: str) -> None:
        self.resource_name = resource_name
        super().__init__(f"Resource '{resource_name}' not defined!")


def get_builder_class(primitive):
    return getattr(sys.modules[__name__], f"{primitive.__class__.__name__}Builder")


def prepare_name(name: str) -> str:
    return "".join([s.capitalize() for s in name.replace("_", "-").split("-")])


class AppStack(Stack):
    def __init__(self, scope: Construct, app: Application, **kwargs) -> None:
        self.app_ = app
        self.built_resources_ = {}
        self.required_layers = []
        self.config = Configuration.get_configuration()

        super().__init__(scope, app.name, env=self.config.env, **kwargs)

        self._build_layers()
        self._build_resources()
        self._build_handlers()

    def _build_resources(self):
        for resource in self.app_.resources.values():
            builder_class = getattr(
                sys.modules[__name__], f"{resource.__class__.__name__}Builder"
            )
            # FIXME: raise error if there were resources with identical names
            aws_resource = builder_class(self, resource).build()
            self._register_resource(resource.name, aws_resource)

    def _register_resource(self, name: str, resource: AwsResource) -> None:
        """Register resource in application

        Args:
            name (str): resource unique name
            resource (AwsResource): construct that represent aws resource

        Raises:
            ResourceAlreadyRegistered: if resource with such name already registered
        """
        if name in self.built_resources_:
            raise ResourceAlreadyRegistered(name)
        self.built_resources_[name] = resource

    def _build_handlers(self):
        for handler in self.app_.handlers:
            handler_class = getattr(
                sys.modules[__name__], f"{handler.__class__.__name__}Builder"
            )
            handler_class(self, handler).build()

    def _build_layers(self):
        lib_folder = Path("./.layers")
        if not lib_folder.exists():
            os.mkdir(str(lib_folder))

        if Path("./shared").exists():
            self._prepare_shared_layer()
            self._build_shared_layer()
        self._prepare_libraries_layer()
        self._build_lib_layer()

    def _prepare_shared_layer(self):
        logging.info("Preparing shared layer")
        shared_layer_folder = Path("./.layers/shared")
        if os.path.exists(shared_layer_folder):
            shutil.rmtree(shared_layer_folder)
        os.mkdir(shared_layer_folder)
        shutil.copytree("shared", str(shared_layer_folder.joinpath("shared")))

    def _prepare_libraries_layer(self):
        # TODO: create requirements only if requirements.txt was changed
        logging.info("Preparing libraries layer")
        lib_layer_folder = Path("./.layers/lib")
        if os.path.exists(lib_layer_folder):
            shutil.rmtree(lib_layer_folder)
        os.mkdir(lib_layer_folder)
        shutil.copy("requirements.txt", str(lib_layer_folder))

    def _build_shared_layer(self):
        self._shared_layer = aws_lambda_python_alpha.PythonLayerVersion(
            self,
            "SharedLayer",
            entry=str(Path("./.layers/shared")),
            compatible_runtimes=[self.config.python_version],
        )
        self.required_layers.append(self._shared_layer)

    def _build_lib_layer(self):
        self._lib_layer = aws_lambda_python_alpha.PythonLayerVersion(
            self,
            "LibLayer",
            entry=str(Path("./.layers/lib")),
            compatible_runtimes=[self.config.python_version],
        )
        self.required_layers.append(self._lib_layer)

    def _build_sqs(self, sqs: Sqs):
        queue = aws_sqs.Queue(
            self, sqs.name, visibility_timeout=Duration.seconds(sqs.visibility_timeout)
        )
        self.built_resources_[sqs.name] = queue
        return queue

    def get_built_resource(self, name: str):
        if name not in self.built_resources_:
            raise ResourceNotDefined(name)
        return self.built_resources_[name]


# ______________ Handler Builders __________________ #


HandlerType = TypeVar("HandlerType", bound=Handler)


class HandlerBuilder(Generic[HandlerType]):
    def __init__(self, context: "AppStack", handler: HandlerType) -> None:
        self.context = context
        self.handler = handler

    def _build_lambda(self):
        dir_ = Path(inspect.getfile(self.handler.func)).parent

        # TODO: add more config options
        lambda_fn = aws_lambda.Function(
            self.context,
            prepare_name(self.handler.name),
            **{
                **{
                    "runtime": self.context.config.python_version,
                    "handler": f"handler.{self.handler.func.__name__}",
                    "code": aws_lambda.Code.from_asset(str(dir_)),
                    "environment": {
                        "APP_NAME": self.context.app_.name,
                    },
                    "layers": self.context.required_layers,
                },
                **self._get_vpc_config(),
                **self._get_handler_config(),
            },
        )

        return lambda_fn

    def _get_vpc_config(self) -> dict:
        # TODO: add vpc config only if we allow it for lambda
        vpc_config = {}
        if self.context.config.vpc_id:
            vpc_config.update(
                {
                    "vpc": aws_ec2.Vpc.from_lookup(
                        self.context,
                        self.context.config.vpc_id,
                        vpc_id=self.context.config.vpc_id,
                    ),
                    "vpc_subnets": aws_ec2.SubnetSelection(
                        subnets=[
                            aws_ec2.Subnet.from_subnet_id(self.context, s_id, s_id)
                            for s_id in self.context.config.subnets_ids  # type: ignore
                        ]
                    ),
                    "security_groups": [
                        aws_ec2.SecurityGroup.from_lookup_by_id(
                            self.context,
                            self.context.config.security_group_id,  # type: ignore
                            self.context.config.security_group_id,  # type: ignore
                        )
                    ],
                    "allow_public_subnet": self.context.config.allow_public_subnet,
                }
            )
        return vpc_config

    def _get_handler_config(self) -> dict:
        return {
            "reserved_concurrent_executions": self.handler.reserved_concurrent_executions
            or self.context.config.lambda_defaults.reserved_concurrent_executions,
            "memory_size": self.handler.memory
            or self.context.config.lambda_defaults.memory_size,
            "timeout": Duration.seconds(
                self.handler.timeout or self.context.config.lambda_defaults.timeout
            ),
            "environment": {
                **self.context.config.lambda_defaults.env_variables,
                **self.handler.env_variables,
            },
        }

    def _connect_resources(self, lambda_: aws_lambda.Function):
        # FIXME: very bad pattern
        for connector in self.handler.resources:
            get_builder_class(connector)(self.context, connector, lambda_).build()

    def build(self):
        # TODO: rework inheritance
        lambda_ = self._build_lambda()
        self._connect_resources(lambda_)
        return lambda_


class ApiHandlerBuilder(HandlerBuilder[ApiHandler]):
    api_ = None

    @property
    def api(self) -> aws_apigateway.RestApi:
        if not self.__class__.api_:
            self.__class__.api_ = aws_apigateway.RestApi(
                self.context,
                f"RestApi",
                deploy_options=aws_apigateway.StageOptions(
                    stage_name=self.context.config.env_name
                ),
            )
        return self.__class__.api_

    def build(self):
        lambda_ = super().build()
        self._build_endpoint(lambda_)
        return lambda_

    def _build_endpoint(self, lambda_: aws_lambda.Function):
        # TODO: investigate what is allow_test_invoke
        integration = aws_apigateway.LambdaIntegration(lambda_, allow_test_invoke=False)  # type: ignore

        endpoint = self.api.root
        for path_part in (p for p in self.handler.path.split("/") if p):
            endpoint = endpoint.get_resource(path_part) or endpoint.add_resource(
                path_part
            )
        for method in self.handler.methods:
            endpoint.add_method(method, integration)


class JobHandlerBuilder(HandlerBuilder[JobHandler]):
    def build(self):
        lambda_ = super().build()
        self._build_rule(lambda_)
        return lambda_

    def _build_rule(self, lambda_):
        aws_events.Rule(
            self.context,
            f"{prepare_name(self.handler.name)}Rule",
            schedule=aws_events.Schedule.expression(self.handler.schedule),
            targets=[aws_events_targets.LambdaFunction(lambda_)],
        )


class SqsHandlerBuilder(HandlerBuilder[SqsHandler]):
    def build(self):
        lambda_ = super().build()
        self._handler_connect_queue(lambda_)
        return lambda_

    def _handler_connect_queue(self, lambda_: aws_lambda.Function):
        queue: aws_sqs.Queue = self.context.get_built_resource(self.handler.queue_name)
        # TODO: add more config options including enabling report_batch_item_failure
        _sqs_event_source = aws_lambda_event_sources.SqsEventSource(queue)
        lambda_.add_event_source(_sqs_event_source)


class S3HandlerBuilder(HandlerBuilder[S3Handler]):
    def build(self):
        lambda_ = super().build()
        self._handler_connect_bucket(lambda_)

    def _handler_connect_bucket(self, lambda_: aws_lambda.Function):
        bucket = aws_s3.Bucket = self.context.get_built_resource(
            self.handler.bucket_name
        )
        # FIXME: add events
        events_ = [getattr(aws_s3.EventType, e.name) for e in self.handler.events]
        filters_ = []
        _s3_event_source = aws_lambda_event_sources.S3EventSource(
            bucket, events=events_, filters=filters_
        )
        lambda_.add_event_source(_s3_event_source)


# ______________ Resource Builders __________________ #


ResourceType = TypeVar("ResourceType", bound=Resource)


class ResourceBuilder(ABC, Generic[ResourceType]):
    def __init__(self, context: "AppStack", resource: ResourceType) -> None:
        self.context = context
        self.resource = resource

    @abstractmethod
    def build(self):
        ...


class SqsBuilder(ResourceBuilder[Sqs]):
    def build(self):
        # TODO: add more config options
        queue = aws_sqs.Queue(
            self.context,
            prepare_name(self.resource.name),
            visibility_timeout=Duration.seconds(self.resource.visibility_timeout),
        )
        return queue


class S3Builder(ResourceBuilder[S3]):
    def build(self):
        bucket = aws_s3.Bucket(
            self.context,
            prepare_name(self.resource.name),
            # bucket_name=f"{self.resource.name}-{uuid.uuid4()}",
        )
        return bucket


# _____________________ Resource Connector Builder ________________________

ConnectorType = TypeVar("ConnectorType", bound=ResourceConnector)


class ResourceConnectorBuilder(ABC, Generic[ConnectorType]):
    def __init__(
        self,
        context: "AppStack",
        connector: ConnectorType,
        lambda_: aws_lambda.Function,
    ) -> None:
        self.context = context
        self.connector = connector
        self.lambda_ = lambda_

    @abstractmethod
    def build(self):
        ...


class SqsConnectorBuilder(ResourceConnectorBuilder[SqsConnector]):
    def build(self):
        # TODO: validate resource type
        resource: aws_sqs.Queue = self.context.get_built_resource(
            self.connector.resource_name
        )
        if self.connector.permission in [SqsPermission.read, SqsPermission.full_access]:
            resource.grant_consume_messages(self.lambda_)
        if self.connector.permission in [
            SqsPermission.write,
            SqsPermission.full_access,
        ]:
            resource.grant_send_messages(self.lambda_)
        self.lambda_.add_environment(
            f"{self.connector.resource_name.upper()}_URL", resource.queue_url
        )


class S3ConnectorBuilder(ResourceConnectorBuilder[S3Connector]):
    def build(self):
        bucket: aws_s3.Bucket = self.context.get_built_resource(
            self.connector.resource_name
        )
        if self.connector.permission is S3Permission.read:
            bucket.grant_read(self.lambda_)
        elif self.connector.permission is S3Permission.write:
            # Rework permissions
            bucket.grant_write(self.lambda_)
            bucket.grant_delete(self.lambda_)
        elif self.connector.permission is S3Permission.full_access:
            bucket.grant_read_write(self.lambda_)
            bucket.grant_delete(self.lambda_)
        self.lambda_.add_environment(
            f"{self.connector.resource_name.upper()}_NAME", bucket.bucket_name
        )



sqs_worker_template = """from viburnum.application import sqs_handler, SqsEventsSequence


@sqs_handler("{sqs_name}")
def {handler_name}(events: SqsEventsSequence):
    for e in events:
        print(e.message_id)
        print(e.body)
"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['viburnum', 'viburnum.application', 'viburnum.cli', 'viburnum.deployer']

package_data = \
{'': ['*']}

extras_require = \
{'deployer': ['constructs>=2.51,<3.0',
              'aws-cdk-lib>=2.50.0,<3.0.0',
              'boto3>=1.26.8,<2.0.0',
              'typer[all]>=0.7.0,<0.8.0',
              'aws-cdk.aws-lambda-python-alpha>=2.51.1a0,<3.0.0']}

entry_points = \
{'console_scripts': ['viburnum = viburnum.cli:app']}

setup_kwargs = {
    'name': 'viburnum',
    'version': '0.1.7.dev1',
    'description': "It's abstraction on top of AWS CDK, that helps in building serverless web applications.",
    'long_description': '# Viburnum\n\n**Viburum** - it\'s a small framework built on top of AWS CDK to simplify development and deploying AWS Serverless web applications.\n\n## Installing\n\n### Installing CDK toolkit\n\nViburnum uses [CDK toolkit](https://docs.aws.amazon.com/cdk/v2/guide/cli.html), so first of all you need to install it:\n\n```bash\nnpm install -g aws-cdk  \n```\n\n### Installing Viburnum\n\nPackage consist of two pats `primitives` that help to describe your handlers and resources and `deployer` that will convert primitives into CloudFormation using CDK.\n\n#### Installing only primitives\n\n```bash\npip install viburnum\n```\n\n#### Installing with deployer\n\n```bash\npip install "viburnum[deployer]"\n```\n\nLambda function will require only primitives to work correctly. That\'s why it\'s recommended to add `viburnum` into `requirements.txt` and `viburnum[deployer]` into `requirements-dev.txt`\n\n## Project structure\n\nEach Lambda function handler is represented as folder with `handler.py` inside and other files if required.\n\n**Example** `handler.py`:\n\n```python\nfrom viburnum.application import Request, Response, route\n\n@route("/tests/{id}", methods=["GET"])\ndef get_test(request: Request, test_queue):\n    print(f"Get test: {request.path_params.get(\'id\')}")\n    return Response(200, {})\n```\n\nIn the root folder you need to have `app.py` file with `Application`, this file used by deployer and CDK to determine all related resources.\n\n**Example** `app.py`\n\n```python\nimport aws_cdk as cdk\nfrom viburnum.deployer.builders import AppConstruct\nfrom viburnum.application import Application, Sqs\n\nfrom functions.api.get_test.handler import get_test\n\napp = Application("TestApp")\n# Handlers\napp.add_handler(get_test)\n\ncontext = cdk.App()\nAppConstruct(context, app)\ncontext.synth()\n```\n\nAll logic that shared across all lambdas, must be placed inside `shared` folder, and it will be plugged into Lambda as a Layer.\n\nConfigurations are stored in `config.py` file.\n\n**Example** `config.py`\n\n```python\n# config.py\nfrom viburnum.deployer import Configuration\nfrom aws_cdk import Environment, aws_lambda\n\n\n# DEV CONFIG\nConfiguration(\n    env_name="dev",\n    env=Environment(account="267695621767", region="eu-central-1"),\n    python_version=aws_lambda.Runtime.PYTHON_3_9,\n    ...\n)\n```\n\n### Recommended structure\n\n```project\n├── functions\n│   ├── __init__.py\n│   ├── api\n│   │   ├── __init__.py\n│   │   ├── some_api\n│   │   │    ├── __init__.py\n│   │   │    ├── handler.py\n│   │   │    └── ...\n│   │   │\n│   │   └── ...\n│   │   \n│   ├── jobs\n│   │   ├── __init__.py\n│   │   ├── some_job\n│   │   │    ├── __init__.py\n│   │   │    ├── handler.py\n│   │   │    └── ...\n│   │   │\n│   │   └── ...\n│   │   \n│   └── workers\n│       ├── __init__.py\n│       ├── some_job\n│       │    ├── __init__.py\n│       │    ├── handler.py\n│       │    └── ...\n│       │\n│       └── ...\n│      \n├── shared\n│   ├── __init__.py\n│   └── ...\n│\n├── app.py\n├── config.py\n├── requirements-dev.txt\n└── requirements.txt\n```\n\n### CLI tool\n\nViburnum deployer include CLI tool that helps initializing project and creating a new handlers.\nAfter initializing project folder with `cdk init app --language python` you can call `virburnum init`, that command will change some files so Virburnum can work correctly.\nThere is command for creating new handlers `virburnum add [HANDLER_TYPE]` that will create a handler.\n\nSupported `HANDLER_TYPE`:\n\n- `api`\n- `worker`\n- `job`\n\n## Example app\n\nSimple [example app](https://github.com/yarik2215/Viburnum-example)\n',
    'author': 'Yaroslav Martynenko',
    'author_email': 'stikblacklabel@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/yarik2215/Viburnum',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



app_template = """#!/usr/bin/env python3
import aws_cdk as cdk
from viburnum.deployer import AppStack
from viburnum.application import Application

# [Imports]

app = Application("{app_name}")

# [Handlers]

# [Resources]

cdk_app = cdk.App()
AppStack(cdk_app, app)

cdk_app.synth()
"""
